/*
**  Copyright (C) 1993/1994, OFFIS, Oldenburg University and CERIUM
**  
**  This software and supporting documentation were
**  developed by 
**  
**    Institut OFFIS
**    Bereich Kommunikationssysteme
**    Westerstr. 10-12
**    26121 Oldenburg, Germany
**    
**    Fachbereich Informatik
**    Abteilung Prozessinformatik
**    Carl von Ossietzky Universitaet Oldenburg 
**    Ammerlaender Heerstr. 114-118
**    26111 Oldenburg, Germany
**    
**    CERIUM
**    Laboratoire SIM
**    Faculte de Medecine
**    2 Avenue du Pr. Leon Bernard
**    35043 Rennes Cedex, France
**  
**  for CEN/TC251/WG4 as a contribution to the Radiological 
**  Society of North America (RSNA) 1993 Digital Imaging and 
**  Communications in Medicine (DICOM) Demonstration.
**  
**  THIS SOFTWARE IS MADE AVAILABLE, AS IS, AND NEITHER OFFIS,
**  OLDENBURG UNIVERSITY NOR CERIUM MAKE ANY WARRANTY REGARDING 
**  THE SOFTWARE, ITS PERFORMANCE, ITS MERCHANTABILITY OR 
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER 
**  DISEASES OR ITS CONFORMITY TO ANY SPECIFICATION.  THE 
**  ENTIRE RISK AS TO QUALITY AND PERFORMANCE OF THE SOFTWARE   
**  IS WITH THE USER. 
**  
**  Copyright of the software and supporting documentation
**  is, unless otherwise stated, jointly owned by OFFIS,
**  Oldenburg University and CERIUM and free access is hereby
**  granted as a license to use this software, copy this
**  software and prepare derivative works based upon this
**  software. However, any distribution of this software
**  source code or supporting documentation or derivative
**  works (source code and supporting documentation) must
**  include the three paragraphs of this copyright notice. 
** 
*/
/*
** 
** Author: Didier Lemoine		Created: 20-07-93 
** 
** Module: db_store 
** 
** Purpose: 	
** This file contains the routines which provide store facilities
** for the DB Module
** 
** Module Prefix: DB_ 
** 
** Last Update:		hewett
** Update Date:		1996/09/27 08:46:51
** Source File:		/share/dicom/cvs-depot/dcmtk/imagectn/libsrc/dbstore.cc,v
** CVS/RCS Revision:	1.5
** Status:		Exp
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */

BEGIN_EXTERN_C
#include <stdio.h>
#include <errno.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_FILE_H
#include <sys/file.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
END_EXTERN_C

#include "dcompat.h"
#include "imagedb.h"
#include "dbpriv.h"
#include "dcfilefo.h"
#include "dimse.h"


static BOOLEAN quotaSystemEnabled = TRUE;

void DB_enableQuotaSystem(BOOLEAN enable)
{
    quotaSystemEnabled = enable;
}


/*
** Image file deleting 
*/


CONDITION DB_deleteImageFile(char* imgFile)
{
    int lockfd;

    if (!quotaSystemEnabled) {
	fprintf(stderr, "DB QUOTA: disabled: retaining: %s\n", imgFile);
	return DB_NORMAL;
    } else {
	fprintf(stderr, "DB QUOTA: enabled: deleting: %s\n", imgFile);
    }
	
    lockfd = open(imgFile, O_RDWR, 0666);	/* obtain file descriptor */
    if (lockfd < 0) {
	fprintf(stderr, "DB ERROR: cannot open image file for deleting: %s\n", 
		imgFile);
	return DB_ERROR;
    }
    if (flock(lockfd, LOCK_EX) < 0) {	/* exclusive lock (blocking) */
	fprintf(stderr, "DB ERROR: cannot lock image file  for deleting: %s\n", 
		imgFile);
    }
    unlink(imgFile);			/* delete file */
    if (flock(lockfd, LOCK_UN) < 0) {	/* unlock */
	fprintf(stderr, "DB ERROR: cannot unlock image file  for deleting: %s\n", 
		imgFile);
    }
    close(lockfd);				/* release file descriptor */
	
    return DB_NORMAL;
}

/*************************
**	 Delete oldest study in database  
 */

extern int
DB_DeleteOldestStudy(DB_Private_Handle *phandle, StudyDescRecord *pStudyDesc) 
{
    int oldestStudy ;
    double OldestDate ;
    int s ;
    int n ;
    int idx = 0 ;
    IdxRecord idxRec ;

    oldestStudy = 0 ;
    OldestDate = pStudyDesc[0]. LastRecordedDate ;

#ifdef DEBUG
    DB_debug(1, "DB_DeleteOldestStudy\n") ;
#endif

    for ( s = 1 ; s < phandle -> maxStudiesAllowed ; s++ ) {
	if ( ( pStudyDesc[s]. NumberofRegistratedImages != 0 ) && ( pStudyDesc[s]. LastRecordedDate < OldestDate ) ) {
	    OldestDate = pStudyDesc[s]. LastRecordedDate ;
	    oldestStudy = s ;
	}
    }

#ifdef DEBUG
    DB_debug(1, "DB_DeleteOldestStudy oldestStudy = %d\n",oldestStudy) ;
#endif

    n = strlen(pStudyDesc[oldestStudy].StudyInstanceUID) ;
    while ( DB_IdxRead (phandle, idx, &idxRec) == DB_NORMAL ) {

	if ( ! ( strncmp(idxRec. StudyInstanceUID, pStudyDesc[oldestStudy].StudyInstanceUID, n) ) ) {
	    DB_IdxRemove (phandle, idx) ;
	    DB_deleteImageFile(idxRec.filename);
	}
	idx++ ;
    }
	
    pStudyDesc[oldestStudy].NumberofRegistratedImages = 0 ;
    pStudyDesc[oldestStudy].StudySize = 0 ;
    return(oldestStudy) ;
}




/*************************
**	 Delete oldest images in database  
 */

CONDITION
DB_DeleteOldestImages(DB_Private_Handle *phandle, StudyDescRecord *pStudyDesc, int StudyNum, char *StudyUID, long RequiredSize) 
{

    typedef int (*SortProc) (const void *, const void *) ;
 
    ImagesofStudyArray *StudyArray ;
    IdxRecord idxRec ;
    int nbimages = 0 , s = 0 , n ; 
    long DeletedSize ; 

#ifdef DEBUG
    DB_debug(1, "DB_DeleteOldestImages RequiredSize = %ld\n",RequiredSize) ;
#endif
    n = strlen(StudyUID) ;
    StudyArray = (ImagesofStudyArray *)malloc(MAX_NUMBER_OF_IMAGES * sizeof(ImagesofStudyArray)) ;

    if (StudyArray == NULL) {
	fprintf(stderr, "DB_DeleteOldestImages: out of memory\n");
	return DB_ERROR;
    }

    /** Find all images having the same StudyUID
     */

    DB_IdxInitLoop (phandle, &(phandle -> idxCounter)) ;
    while (	DB_IdxGetNext(phandle, &(phandle -> idxCounter), &idxRec) == DB_NORMAL ) {
	if ( ! ( strncmp(idxRec. StudyInstanceUID, StudyUID, n) ) ) {

	    StudyArray[nbimages]. idxCounter = phandle -> idxCounter ;
	    StudyArray[nbimages]. RecordedDate = idxRec. RecordedDate ; 
	    StudyArray[nbimages++]. ImageSize = idxRec. ImageSize ;
	}
    }

    /** Sort the StudyArray in order to have the oldest images first
     */
    qsort((char *)StudyArray, nbimages, sizeof(ImagesofStudyArray), (SortProc) DB_Compare) ;

#ifdef DEBUG
    {
	int i ;
	DB_debug(1, "DB_DeleteOldestImages : Sorted images ref array\n") ;
	for (i = 0 ; i < nbimages ; i++)
	    DB_debug(1, "[%2d] :   Size %ld   Date %20.3f   Ref %d \n",
		    i, StudyArray[i]. ImageSize, StudyArray[i]. RecordedDate, StudyArray[i]. idxCounter) ;
	DB_debug(1, "DB_DeleteOldestImages : end of ref array\n") ;
    }
#endif

    s = 0 ;
    DeletedSize = 0 ;

    while ( DeletedSize < RequiredSize ) {

	IdxRecord idxRemoveRec ;
	DB_IdxRead (phandle, StudyArray[s]. idxCounter, &idxRemoveRec) ;
#ifdef DEBUG
	DB_debug(1, "Removing file : %s\n", idxRemoveRec. filename) ;
#endif
	DB_deleteImageFile(idxRemoveRec.filename);

	DB_IdxRemove (phandle, StudyArray[s]. idxCounter) ;
	pStudyDesc[StudyNum].NumberofRegistratedImages -= 1 ;
	pStudyDesc[StudyNum].StudySize -= StudyArray[s]. ImageSize ;
	DeletedSize += StudyArray[s++]. ImageSize ;
    }

#ifdef DEBUG
    DB_debug(1, "DB_DeleteOldestImages DeletedSize = %d\n",
	     (int)DeletedSize) ;
#endif
    free(StudyArray) ;
    return(	DB_NORMAL ) ;	

}



/*************************
**	 Verify if study UID already exists  
 */

extern int
DB_MatchStudyUIDInStudyDesc (StudyDescRecord *pStudyDesc, char *StudyUID, int maxStudiesAllowed)
{
    int s = 0 ;
    int n ;

    n = strlen(StudyUID) ;

    while ( ( pStudyDesc[s]. NumberofRegistratedImages != 0 ) && ( s < maxStudiesAllowed ) ) {
	if ( ! ( strncmp(pStudyDesc[s]. StudyInstanceUID, StudyUID, n) ) )
	    break ;
	s++ ;
    } ;
		
    return (s) ;
}


/*************************
**	Check up storage rights in Study Desk record 
 */

CONDITION
DB_CheckupinStudyDesc(DB_Private_Handle *phandle, StudyDescRecord *pStudyDesc, char *StudyUID, long imageSize)
{
    int 		s ;
    long 		RequiredSize ;

    s = DB_MatchStudyUIDInStudyDesc (pStudyDesc, StudyUID, 
				     phandle -> maxStudiesAllowed) ;

    /** If Study already exists
     */

    if ( pStudyDesc[s]. NumberofRegistratedImages != 0 ) {

#ifdef DEBUG
	DB_debug(1, "DB_CheckupinStudyDesc: study already exists : %d\n",s) ;
#endif
	if ( ( pStudyDesc[s]. StudySize + imageSize ) 
	     > phandle -> maxBytesPerStudy ) {
	    if ( imageSize > phandle -> maxBytesPerStudy ) {
#ifdef DEBUG
		DB_debug(1, 
			 "DB_CheckupinStudyDesc: imageSize = %ld too large\n",
			 imageSize) ;
#endif
		return ( DB_ERROR ) ;
	    }

	    RequiredSize = imageSize - 
		( phandle -> maxBytesPerStudy - pStudyDesc[s]. StudySize ) ;
	    DB_DeleteOldestImages(phandle, pStudyDesc, s, 
				  StudyUID, RequiredSize) ;
	}


    }
    else {
#ifdef DEBUG
	DB_debug(1, "DB_CheckupinStudyDesc: study doesn't already exist\n") ;
#endif
	if ( imageSize > phandle -> maxBytesPerStudy ) {
#ifdef DEBUG
	    DB_debug(1, "DB_CheckupinStudyDesc: imageSize = %ld too large\n",
		     imageSize) ;
#endif
	    return ( DB_ERROR ) ;
	}	
	if ( s > ( phandle -> maxStudiesAllowed - 1 ) )
	    s = DB_DeleteOldestStudy(phandle, pStudyDesc) ;

    }

    pStudyDesc[s]. StudySize += imageSize ;
#ifdef DEBUG
    DB_debug(1, "DB_CheckupinStudyDesc: ~~~~~~~~ StudySize = %ld\n",
	     pStudyDesc[s]. StudySize) ;
#endif

    /* we only have second accuracy */
    pStudyDesc[s]. LastRecordedDate =  (double) time(NULL);

    pStudyDesc[s]. NumberofRegistratedImages++ ;
    strcpy(pStudyDesc[s].StudyInstanceUID,StudyUID) ;

    if ( DB_StudyDescChange (phandle, pStudyDesc) == DB_NORMAL)
	return ( DB_NORMAL ) ;
    else 
	return ( DB_ERROR ) ;


}

/*
 * If the image is already stored remove it from the database.
 * hewett - Nov. 1, 93
 */
CONDITION DB_removeDuplicateImage(DB_Private_Handle *phandle, 
	const char *SOPInstanceUID, const char *StudyInstanceUID, 
	StudyDescRecord *pStudyDesc, const char *newImageFileName)
{
    int idx = 0;
    IdxRecord 		idxRec ;
    int studyIdx = 0;

    studyIdx = DB_MatchStudyUIDInStudyDesc (pStudyDesc, (char*)StudyInstanceUID, 
					    phandle -> maxStudiesAllowed) ;

    if ( pStudyDesc[studyIdx].NumberofRegistratedImages == 0 ) {
	/* no study images, cannot be any old images */
	return DB_NORMAL;
    }
	
    while (DB_IdxRead(phandle, idx, &idxRec) == DB_NORMAL) {

	if (strcmp(idxRec.SOPInstanceUID, SOPInstanceUID) == 0) {

#ifdef DEBUG
	    DB_debug(1,"--- Removing Existing DB Image Record: %s\n", 
		   idxRec.filename) ;
#endif
	    /* remove the idx record  */
	    DB_IdxRemove (phandle, idx);
	    /* only remove the image file if it is different than that
	     * being entered into the database.
	     */
	    if (strcmp(idxRec.filename, newImageFileName) != 0) {
		DB_deleteImageFile(idxRec.filename);
	    }
	    /* update the study info */
	    pStudyDesc[studyIdx].NumberofRegistratedImages--;
	    pStudyDesc[studyIdx].StudySize -= idxRec.ImageSize;
	}
	idx++;
    }
    /* the study record should be written to file later */
    return DB_NORMAL;
}


/*************************
**	Add data from imageFileName to database
 */

CONDITION 
DB_storeRequest (
	DB_Handle 	*handle, 
	const char 	*SOPClassUID, 
	const char 	* /*SOPInstanceUID*/, 
	const char 	*imageFileName, 
	DB_Status 	*status) 
{
    DB_Private_Handle 	*phandle ; 
    IdxRecord 		idxRec ;
    StudyDescRecord 	*pStudyDesc ;	
    int 			i ;
    struct stat 		buf ;


    /**** Initialize an IdxRecord
    ***/

    bzero((char*)&idxRec, sizeof(idxRec));

    DB_IdxInitRecord (&idxRec, 0) ;

    strncpy(idxRec.filename, imageFileName, DBC_MAXSTRING);
#ifdef DEBUG
    DB_debug(1,"DB_storeRequest () : storage request of file : %s\n",idxRec . filename) ;
#endif
    strncpy (idxRec.SOPClassUID, SOPClassUID, UI_MAX_LENGTH);

    /**** Get IdxRec values from ImageFile
    ***/

    DcmFileStream inf(imageFileName, DCM_ReadMode);
    if ( inf.Fail() ) {
	fprintf(stderr, "DB: Cannot open file: %s: %s", 
		imageFileName, strerror(errno));
	status->status = STATUS_STORE_Error_CannotUnderstand ;
	return (DB_ERROR) ;
    }

    DcmFileFormat dcmff;
    dcmff.transferInit();
    dcmff.read(inf, EXS_Unknown);
    dcmff.transferEnd();

    if (dcmff.error() != EC_Normal) {
	fprintf(stderr, "DB: Bad DICOM file: %s: %s", imageFileName, 
	       dcmErrorConditionToString(dcmff.error()));
	status->status = STATUS_STORE_Error_CannotUnderstand ;
	return (DB_ERROR) ;
    }
    
    DcmDataset *dset = dcmff.getDataset();

    for (i = 0 ; i < NBPARAMETERS ; i++ ) {
	E_Condition ec = EC_Normal;
	DB_SmallDcmElmt *se = idxRec.param + i;
	
	ec = dset->findString(se->XTag, se->PValueField,
			      se->ValueLength);
	if (ec != EC_Normal) {
	    /* not found */
	    se->ValueLength = 0;
	} else {
	    /* value has already been filled in */
	    se->ValueLength = strlen(se->PValueField);
	}
    }

    /**** Print Elements
    ***/

#ifdef DEBUG
    DB_debug(1,"-- BEGIN Parameters to Register in DB\n") ;
    for (i = 0 ; i < NBPARAMETERS ; i++) {	/* new definition */
	DB_SmallDcmElmt *se = idxRec.param + i;
	char* value = "";
	if (se->PValueField != NULL) value = se->PValueField;
	DcmTag tag(se->XTag);
	DB_debug(1," %s: \"%s\"\n", tag.getTagName(), value);
    }
    DB_debug(1,"-- END Parameters to Register in DB\n") ;
#endif

    /**** Goto the end of IndexFile, and write the record
    ***/

    phandle = (DB_Private_Handle *) handle ; 

    DB_lock(phandle, TRUE);

    pStudyDesc = (StudyDescRecord *)malloc (SIZEOF_STUDYDESC) ;
    if (pStudyDesc == NULL) {
	fprintf(stderr, "DB_storeRequest: out of memory\n");
	status -> status = STATUS_STORE_Refused_OutOfResources ;

	DB_unlock(phandle);

	return (DB_ERROR) ;
    }

    for ( i = 0 ; i < phandle -> maxStudiesAllowed ; i++ )
	pStudyDesc[i]. NumberofRegistratedImages = 0 ;

    DB_GetStudyDesc(phandle, pStudyDesc) ;

    stat(imageFileName, &buf) ;
    idxRec. ImageSize = (int)(buf. st_size) ;

    /* we only have second accuracy */
    idxRec. RecordedDate =  (double) time(NULL);

    /*
     * If the image is already stored remove it from the database.
     * hewett - Nov. 1, 93
     */
    DB_removeDuplicateImage(phandle, idxRec.SOPInstanceUID, 
			    idxRec.StudyInstanceUID, pStudyDesc,
			    imageFileName);


    if ( DB_CheckupinStudyDesc(phandle, pStudyDesc, idxRec. StudyInstanceUID, idxRec. ImageSize) != DB_NORMAL ) {
	free (pStudyDesc) ;
	status -> status = STATUS_STORE_Refused_OutOfResources ;

	DB_unlock(phandle);

	return (DB_ERROR) ;
    }
		
    free (pStudyDesc) ;

    if (DB_IdxAdd (phandle, &i, &idxRec) == DB_NORMAL) {
	status -> status = STATUS_Success ;

	DB_unlock(phandle);

	return (DB_NORMAL) ;
    }
    else {
	status -> status = STATUS_STORE_Refused_OutOfResources ;

	DB_unlock(phandle);

	return (DB_ERROR) ;
    }

    return DB_ERROR;
}

/*
** Prune invalid DB records.
*/

CONDITION DB_pruneInvalidRecords(DB_Handle *dbHandle)
{
    DB_Private_Handle *phandle = (DB_Private_Handle *)dbHandle;
    int idx = 0;
    IdxRecord idxRec ;
    StudyDescRecord *pStudyDesc;

    DB_lock(phandle, TRUE);

    pStudyDesc = (StudyDescRecord *)malloc (SIZEOF_STUDYDESC) ;
    if (pStudyDesc == NULL) {
	fprintf(stderr, "DB_pruneInvalidRecords: out of memory\n");
	DB_unlock(phandle);
	return (DB_ERROR) ;
    }

    for (int i = 0 ; i < phandle -> maxStudiesAllowed ; i++ )
	pStudyDesc[i]. NumberofRegistratedImages = 0 ;

    DB_GetStudyDesc(phandle, pStudyDesc) ;

    while (DB_IdxRead(phandle, idx, &idxRec) == DB_NORMAL) {

	if (access(idxRec.filename, R_OK) < 0) {

#ifdef DEBUG
	    DB_debug(1,"*** Pruning Invalid DB Image Record: %s\n", 
		   idxRec.filename);
#endif
	    /* update the study info */
	    int studyIdx = DB_MatchStudyUIDInStudyDesc(
		pStudyDesc, idxRec.StudyInstanceUID, 
		phandle->maxStudiesAllowed) ;
	    if (studyIdx < phandle->maxStudiesAllowed) {
		if (pStudyDesc[studyIdx].NumberofRegistratedImages > 0) {
		    pStudyDesc[studyIdx].NumberofRegistratedImages--;
		} else {
		    pStudyDesc[studyIdx].NumberofRegistratedImages = 0;
		    pStudyDesc[studyIdx].StudySize = 0;
		    pStudyDesc[studyIdx].StudyInstanceUID[0] = '\0';
		}
		if (pStudyDesc[studyIdx].StudySize > idxRec.ImageSize) {
		    pStudyDesc[studyIdx].StudySize -= idxRec.ImageSize;
		}
	    }

	    /* remove the idx record  */
	    DB_IdxRemove (phandle, idx);

	}
	idx++;
    }

    DB_StudyDescChange (phandle, pStudyDesc);

    DB_unlock(phandle);

    free (pStudyDesc) ;

    return DB_NORMAL;
}

/*
** CVS Log
** dbstore.cc,v
** Revision 1.5  1996/09/27 08:46:51  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C.
**
** Revision 1.4  1996/05/06 07:39:09  hewett
** Added explicit initialization (bzero) of idx record.  Rearranged
** size of string attributes in idx record.
**
** Revision 1.3  1996/04/29 10:15:48  hewett
** Added global flag to enable/disable quota system from deleting
** image files.  By default this flag is true (images will be
** deleted).  The flag can be set true/false via the D
** B_enableQuotaSystem() function.
**
** Revision 1.2  1996/04/22 11:22:54  hewett
** Added function to prune the db index file of records pointing to
** non-existant image files.
**
** Revision 1.1.1.1  1996/03/28 19:25:00  hewett
** Oldenburg Image CTN Software ported to use the dcmdata C++ toolkit.
**
**
*/
