/*
**
**  Copyright (C) 1996, OFFIS
**
**  This software and supporting documentation were developed by
**
**    Kuratorium OFFIS e.V.
**    Forschungsbereich 2: Kommunikationssysteme
**    Escherweg 2
**    D-26121 Oldenburg, Germany
**
**  for CEN/TC251/WG4 as a contribution to the Computer Assisted Radiology
**  (CAR) 1996 DICOM Demonstration.
**
**  THIS SOFTWARE IS MADE AVAILABLE,  AS IS,  AND OFFIS MAKES NO  WARRANTY
**  REGARDING  THE  SOFTWARE,  ITS  PERFORMANCE,  ITS  MERCHANTABILITY  OR
**  FITNESS FOR ANY PARTICULAR USE, FREEDOM FROM ANY COMPUTER DISEASES  OR
**  ITS CONFORMITY TO ANY SPECIFICATION. THE ENTIRE RISK AS TO QUALITY AND
**  PERFORMANCE OF THE SOFTWARE IS WITH THE USER.
**
**  Copyright of the software  and  supporting  documentation  is,  unless
**  otherwise stated, owned by OFFIS, and free access is hereby granted as
**  a license to  use  this  software,  copy  this  software  and  prepare
**  derivative works based upon this software.  However, any  distribution
**  of this software source code or supporting documentation or derivative
**  works  (source code and  supporting documentation)  must  include  the
**  three paragraphs of this copyright notice.
**
*/
/*
** Service Class Executive (SCE)
** - Worklist Find Service Class Provider
**
** Author: 	Andrew Hewett
**		Kuratorium OFFIS e.V., Germany
** Created:	03/1996
**
** Last Update:		hewett
** Update Date:		1996/09/27 08:54:26
** Source File:		/share/dicom/cvs-depot/dcmtk/wlistctn/apps/scefind.cc,v
** CVS/RCS Revision:	1.6
** Status:		Exp
**
** CVS/RCS Log at end of file
*/

#include "osconfig.h"    /* make sure OS specific configuration is included first */

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
END_EXTERN_C

#include "dicom.h"
#include "wlistctn.h"
#include "dimse.h"
#include "diutil.h"
#include "sce.h"
#include "wrklstdb.h"
#ifndef NODISPLAY
#include "dispuser.h"
#endif
#include "dcdeftag.h"

#include "scefind.h"


typedef struct {
    WorklistDB	*dbHandle;
    WDBStatus	priorStatus;
    DIC_AE	ourAETitle;
} SCE_FindContext;

static BOOL
isaPendingStatus(WDBStatus status)
{
    return ((status == WDB_Pending) || (status == WDB_Pending_Warning));
}

static const char*
status2string(WDBStatus status) 
{
    const char* s = NULL;
    switch (status) {
    case WDB_Refused_OutOfResorces:
	s = "Refused_OutOfResorces"; break;
    case WDB_Failed_IdentifierDoesNotMatchSOPClass:
	s = "Failed_IdentifierDoesNotMatchSOPClass"; break;
    case WDB_Failed_UnableToProcess:
	s = "Failed_UnableToProcess"; break;
    case WDB_Cancel:
	s = "Cancel"; break;
    case WDB_Success:
	s = "Success"; break;
    case WDB_Pending:
	s = "Pending"; break;
    case WDB_Pending_Warning:
	s = "Pending_Warning"; break;
    default:
	s = "<UNKNOWN-STATUS>"; break;
    }
    return s;
}

static void
addStatusDetail(DcmDataset **statusDetail, const DcmElement *elem)
{
    if (elem == NULL) {
	return;
    }

    if (*statusDetail == NULL) {
	*statusDetail = new DcmDataset();
    }

    switch (elem->ident()) {
    case EVR_LO: {
	DcmLongString *lo = new DcmLongString(*((DcmLongString*)elem));
	(*statusDetail)->insert(lo);
	break;
    }
    case EVR_AT: {
	DcmAttributeTag *at = new DcmAttributeTag(*((DcmAttributeTag*)elem));
	(*statusDetail)->insert(at);
	break;
    }
    default: {
	/* other status detail is not supported */
	DcmVR vr(elem->ident());
	errmsg("addStatusDetail: unsupported status detail type: %s",
	       vr.getVRName());
	break;
    }
    }
}

static void
findCallback(
	/* in */ 
	void *callbackData,  
	BOOLEAN cancelled, T_DIMSE_C_FindRQ */*request*/, 
	DcmDataset *requestIdentifiers, int responseCount,
	/* out */
	T_DIMSE_C_FindRSP *response,
	DcmDataset **responseIdentifiers,
	DcmDataset **statusDetail)
{
    WDBStatus dbstatus;
    SCE_FindContext *context;
    WorklistDB* db;

    context = (SCE_FindContext*)callbackData;	/* recover context */
    db = context->dbHandle;

    dbstatus = context->priorStatus;
    
    if (responseCount == 1) {
        /* start the database search */
	if (verbose) {
	    printf("Find SCP Request Identifiers:\n");
	    requestIdentifiers->print();
        }
	dbstatus = db->startFindRequest(*requestIdentifiers);
        if (!isaPendingStatus(dbstatus)) {
	    errmsg(
		"findSCP: Worklist Database: startFindRequest() Failed (%s):",
		status2string(dbstatus));
	}
    }
    
    /* only cancel if we have pending responses */
    if (cancelled && isaPendingStatus(dbstatus)) {
	dbstatus = db->cancelFindRequest();
    }

    if (isaPendingStatus(dbstatus)) {
	/* create the reponse identifiers */
	*responseIdentifiers = new DcmDataset();

	/* get the next response */
	dbstatus = db->nextFindResponse(**responseIdentifiers);
	if (dbstatus != WDB_Success && !isaPendingStatus(dbstatus)) {
	    errmsg(
		"findSCP: Worklist Database: nextFindResponse() Failed (%s):",
		status2string(dbstatus));
	}
	if (debug) {
	    printf("Response Identifiers (%d)\n", responseCount);
	    (*responseIdentifiers)->print();
	    printf("-------\n");
	}
	/* 
	** a pending status will cause the response identifiers to be 
	** deleted by the calling function.
	*/
	if (!isaPendingStatus(dbstatus)) {
	    /* logic check */
	    if ((*responseIdentifiers)->card() > 0) {
		errmsg("ERROR: non-pending status with a response!");
	    }
	    /* delete the response identifers */
	    delete *responseIdentifiers;
	    *responseIdentifiers = NULL;
	}
    }

    if (verbose) {
        printf("Worklist Find SCP Response %d [status: %s]\n", responseCount,
	    status2string(dbstatus));
    }

    /* set response status */
    response->Status = dbstatus;

    if (*statusDetail != NULL) {
	delete *statusDetail;
	*statusDetail = NULL;
    }

    /*
    ** Status detail may only be returned for particular status codes.
    */
    switch (dbstatus) {
    case WDB_Failed_IdentifierDoesNotMatchSOPClass:
    case WDB_Failed_UnableToProcess:
	addStatusDetail(statusDetail, db->getOffendingElements());
	addStatusDetail(statusDetail, db->getErrorComments());
	break;
    case WDB_Refused_OutOfResorces:
	/* out of resources may only have error comment detail */
	addStatusDetail(statusDetail, db->getErrorComments());
	break;
    default:
	/* other status codes may not have any status detail */
	break;
    }
}


CONDITION
SCE_findSCP(T_ASC_Association * assoc, T_DIMSE_C_FindRQ * request,
	T_ASC_PresentationContextID presID,
	WorklistDB &dbHandle)

{
    CONDITION cond;
    SCE_FindContext context;

    context.dbHandle = &dbHandle;
    context.priorStatus = WDB_Pending;
    ASC_getAPTitles(assoc->params, NULL, context.ourAETitle, NULL);

    if (verbose) {
	printf("Received Find SCP: ");
	DIMSE_printCFindRQ(stdout, request);
    }

    cond = DIMSE_findProvider(assoc, presID, request, 
    	findCallback, &context, DIMSE_BLOCKING, 0);
    if (!SUCCESS(cond)) {
        errmsg("Find SCP Failed:");
	COND_DumpConditions();
    }
    return cond; 
}


/*
** CVS Log
** scefind.cc,v
** Revision 1.6  1996/09/27 08:54:26  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C
**
** Revision 1.5  1996/08/29 16:16:08  hewett
** Corrected copyright header.
**
** Revision 1.4  1996/05/01 17:33:49  hewett
** Modified for latest Worklist DB.
**
** Revision 1.3  1996/04/29 19:17:23  hewett
** Updated for existance of WorklistDB::getErrorComments().
**
** Revision 1.2  1996/04/27 13:02:28  hewett
** Now only returns status detail together with status codes which
** are allowed to have status detail (see Supplement 10, Table Z.4-1
** C-FIND Response Status Values).
**
** Revision 1.1.1.1  1996/04/26 14:16:01  hewett
** Initial Import.
**
**
**
*/
