/*
**  Copyright (C) 1996, Kuratorium OFFIS e.V., Oldenburg, Germany
*/
/*
** Worklist Database Test Program
**
** Author: 	Andrew Hewett
**		Kuratorium OFFIS e.V., Oldenburg, Germany
** Created:	04/1996
**
** Last Update:		hewett
** Update Date:		1996/09/27 08:54:35
** Source File:		/share/dicom/cvs-depot/dcmtk/wlistctn/tests/wltest.cc,v
** CVS/RCS Revision:	1.5
** Status:		Exp
**
** CVS/RCS Log at end of file
*/


#include "osconfig.h"    /* make sure OS specific configuration is included first */

BEGIN_EXTERN_C
#include <stdio.h>
#include <string.h>
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <errno.h>
END_EXTERN_C

#include "dicom.h"
#include "wrklstdb.h"
#include "dcfilefo.h"
#include "dcdebug.h"
#include "dcdict.h"
#include "cmdlnarg.h"

char* progname = NULL;
BOOL verbose = FALSE;
int debug = 0;

static void
shortusage()
{
    fprintf(stderr, 
	    "usage: %s [-k key][-v][-d<n>] <worklist-db-dir> <query> [<query> ...]\n", 
	    progname);
}

static void
fullusage()
{
    shortusage();
    fprintf(stderr, "\
parameters:\n\
    <worklist-db-dir>	the worklist database directory\n\
    <query>		a dicom file containing a worklist query\n\
    port	tcp/ip port number to listen on\n\
options:\n\
    -k key      override query key (gggg,eeee=\"string\")\n\
    -v		verbose mode\n\
    -d<n>	set debug level <n>\n");
    exit(1);
}

static void 
usage()
{
    shortusage();
    exit(1);
}

void errmsg(const char* msg, ...)
{
    va_list args;
	
    fprintf(stderr, "%s: ", progname);
    va_start(args, msg);
    vfprintf(stderr, msg, args);
    va_end(args);
    fprintf(stderr, "\n");
}


static void
addOverrideKey(DcmDataset& overrideKeys, char* s)
{
    unsigned int g = 0xffff;
    unsigned int e = 0xffff;
    int n = 0;
    char val[1024];

    val[0] = '\0';
    n = sscanf(s, "%x,%x=%s", &g, &e, val);
    
    if (n < 2) {
	errmsg("bad key format: %s", s);
	usage(); /* does not return */
    }

    DcmTag tag(g,e);
    if (tag.error() != EC_Normal) {
	errmsg("unknown tag: (%04x,%04x)", g, e);
	usage();
    }
    DcmElement *elem = newDicomElement(tag);
    if (elem == NULL) {
	errmsg("cannot create element for tag: (%04x,%04x)", g, e);
	usage();
    }
    if (strlen(val) > 0) {
	elem->put(val);
	if (elem->error() != EC_Normal) {
	    errmsg("cannot put tag value: (%04x,%04x)=\"%s\"", g, e, val);
	    usage();
	}
    }

    overrideKeys.insert(elem, TRUE);
    if (overrideKeys.error() != EC_Normal) {
	errmsg("cannot insert tag: (%04x,%04x)", g, e);
	usage();
    }
}

static BOOL
fileExists(const char* fname) 
{
    BOOL ok = FALSE;
    FILE* f = NULL;

    if ((f = fopen(fname, "r")) != NULL) {
	ok = TRUE;
	fclose(f);
    }
    return ok;
}

static BOOL
queryWorklistDB(WorklistDB& wdb, 
		const char* queryFilename, DcmDataset& overrideKeys);


int main(int argc, char* argv[])
{
    int i = 0;
    int j = 0;

#ifdef HAVE_GUSI_H
    /* needed for Macintosh */
    GUSISetup(GUSIwithSIOUXSockets);
    GUSISetup(GUSIwithInternetSockets);
#endif

    prepareCmdLineArgs(argc, argv, "wltest");

    /* strip any leading path from program name */
    if ((progname = (char *)strrchr(argv[0], PATH_SEPARATOR)) != NULL) {
	progname++;
    } else {
	progname = argv[0];
    }

    if (argc < 2) {
	fullusage();
    }

    /* make sure data dictionary is loaded */
    if (dcmDataDict.numberOfEntries() == 0) {
	errmsg("Warning: no data dictionary loaded, check environment variable: %s\n",
		DCM_DICT_ENVIRONMENT_VARIABLE);
    }

    /* a set of query keys to be added/substituted in each query */
    DcmDataset overrideKeys;

    /* parse program arguments */
    for (i = 1; i < argc && argv[i][0] == '-'; i++) {
	switch (argv[i][1]) {
	case 'v':
	    verbose = TRUE;
	    break;
	case 'd':
	    if (((i + 1) < argc) && (argv[i + 1][0] != '-') &&
		(sscanf(argv[i + 1], "%d", &debug) == 1)) {
		i++;		
	    } else {
		debug = 1;
	    }
	    verbose = TRUE;
	    break;
        case 'k':
	    if ((i + 1) < argc) {
		addOverrideKey(overrideKeys, argv[++i]);
	    } else {
		usage();
	    }
	    break;
	default:
	    fullusage();
	}
    }

#ifdef USE_MALLOC_DEBUG
#ifdef HAVE_MALLOC_DEBUG
#define MALLOCDEBUGLEVEL 16
    if (verbose) {
        printf("malloc_debug on (%d)\n", MALLOCDEBUGLEVEL);
    }
    malloc_debug(MALLOCDEBUGLEVEL);
#endif
#endif

    for (j=i; j<argc; j++) {
	if (!fileExists(argv[j])) {
	    errmsg("cannot access %s: %s", argv[j], 
		   strerror(errno));
	    usage();
	}
    }

    SetDebugLevel((debug));	/* dcmdata debugging */

    char* dbPath = argv[i];
    i++;

    WorklistDB wdb(dbPath);

    for (j=i; j<argc; j++) {
	char* queryFilename = argv[j];

	queryWorklistDB(wdb, queryFilename, overrideKeys);
    }

    return 0;
}

static void
substituteOverrideKeys(DcmDataset* query, DcmDataset& overrideKeys)
{
    if (overrideKeys.card() == 0) {
	return; /* nothing to do */
    }

    /* copy the override keys */
    DcmDataset keys(overrideKeys);

    /* put the override keys into query replacing existing tags */
    int elemCount = keys.card();
    for (int i=0; i<elemCount; i++) {
	DcmElement *elem = keys.remove((unsigned long)0);

	query->insert(elem, TRUE);
    }
}

static BOOL
isaPendingStatus(WDBStatus status)
{
    return ((status == WDB_Pending) || (status == WDB_Pending_Warning));
}

static const char*
status2string(WDBStatus status) 
{
    const char* s = NULL;
    switch (status) {
    case WDB_Refused_OutOfResorces:
	s = "Refused_OutOfResorces"; break;
    case WDB_Failed_IdentifierDoesNotMatchSOPClass:
	s = "Failed_IdentifierDoesNotMatchSOPClass"; break;
    case WDB_Failed_UnableToProcess:
	s = "Failed_UnableToProcess"; break;
    case WDB_Cancel:
	s = "Cancel"; break;
    case WDB_Success:
	s = "Success"; break;
    case WDB_Pending:
	s = "Pending"; break;
    case WDB_Pending_Warning:
	s = "Pending_Warning"; break;
    default:
	s = "<UNKNOWN-STATUS>"; break;
    }
    return s;
}

static BOOL
queryWorklistDB(WorklistDB& wdb, 
		const char* queryFilename, DcmDataset& overrideKeys)
{
    
    DcmFileFormat dcmff;

    if (queryFilename != NULL) {
	DcmFileStream inf(queryFilename, DCM_ReadMode);
	if ( inf.Fail() ) {
	    errmsg("Cannot open file: %s: %s", queryFilename, strerror(errno));
	    return FALSE;
	}

	dcmff.transferInit();
	dcmff.read(inf, EXS_Unknown);
	dcmff.transferEnd();

	if (dcmff.error() != EC_Normal) {
	    errmsg("Bad DICOM file: %s: %s", queryFilename, 
		   dcmErrorConditionToString(dcmff.error()));
	    return FALSE;
	}
    }

    /* replace specific keys by those in overrideKeys */
    substituteOverrideKeys(dcmff.getDataset(), overrideKeys);

    if (verbose) {
	printf("------------\n");
	printf("Query:\n");
	dcmff.getDataset()->print();
	printf("------------\n");
	printf("Starting Query\n");
    }

    /* perform the query */
    
    WDBStatus status = WDB_Pending;
    int responseCount = 0;

    /*
    ** Start the find request.
    */
    status = wdb.startFindRequest(*dcmff.getDataset());

    while (isaPendingStatus(status)) {
	
	DcmDataset* response = new DcmDataset;

	/*
	** Get the next find response if there is one.
	*/
	status = wdb.nextFindResponse(*response);
	responseCount++;

	if (isaPendingStatus(status)) {
	    printf("------------\n");
	    printf("Response %d (%s)\n", responseCount, status2string(status));
	    response->print();
	} else if (response->card() > 0) {
	    errmsg("ERROR: non-pending status with a response!");
	}
	delete response;
    }

    if (verbose) {
	printf("------------\n");
	printf("Query Complete (%s)\n", status2string(status));
	if (wdb.getErrorComments() != NULL) {
	    printf("Error Comment:\n");
	    DcmLongString errorComment(*wdb.getErrorComments());
	    errorComment.print();
	}
	if (wdb.getOffendingElements() != NULL) {
	    printf("Offending Elements:\n");
	    DcmAttributeTag offendingElements(*wdb.getOffendingElements());
	    offendingElements.print();
	}
    }

    return TRUE;
}

/*
** CVS Log
** wltest.cc,v
** Revision 1.5  1996/09/27 08:54:35  hewett
** Enclosed system include files with BEGIN_EXTERN_C/END_EXTERN_C
**
** Revision 1.4  1996/09/24 15:50:08  hewett
** Added conditionaly compiled support for the Macintosh
** environment (GUSI library).
**
** Revision 1.3  1996/05/03 10:33:40  hewett
** Added check for existance of file arguments before processing starts.
**
** Revision 1.2  1996/04/29 19:16:23  hewett
** Corrected logic when printing response identifiers.
**
** Revision 1.1.1.1  1996/04/26 14:16:02  hewett
** Initial Import.
**
**
*/
